#lang racket/base

(require racket/contract)

(provide
  (contract-out
    [make-progress-data (-> progress-data?)]
    [add-progress-record!
      (-> progress-data? integer? integer? (or/c integer? BEFORE-GUESSES) any)]
    [progress-data? (-> any/c boolean?)]
    [progress-data->string (-> progress-data? string?)]
    [progress-data->text-file (-> progress-data? output-port? any)]
    [progress-data->node (-> progress-data? node?)]
    [recursion-stats (-> progress-data? (vectorof integer?))]
    [BEFORE-GUESSES BEFORE-GUESSES]
))

(require
  data/gvector
  ffi/vector
  racket/list
  racket/format
  racket/port
  racket/string
  "graph.rkt"
)

; --------------------------------------------------------------------------------
; General handling of progress data

; Denote that we entered the recursion level "from the top", before making any
; guesses. This value must be an uint8 to fit in the `u8vector` record, and the
; value must not be any valid guess from 1 to 9. Otherwise the value is arbitrary.
(define BEFORE-GUESSES 'before-guesses)

(struct record (recursion-level board-index guess)
  ; Make transparent for tests
  #:transparent
  #:constructor-name Record)

; Integer Integer Integer -> U8vector
(define (make-record recursion-level board-index guess)
  (Record recursion-level board-index guess))

; [Gvector-of U8vector]
(struct progress-data (records)
  #:constructor-name Progress-Data)

; Progress-Data Integer Integer Integer -> Void
;
; Add information on the current solver guess. `index` is the index into the
; Sudoku board.
(define (add-progress-record! progress-data recursion-level board-index guess)
  (define records (progress-data-records progress-data))
  (define new-record (make-record recursion-level board-index guess))
  (gvector-add! records new-record))

; [Optional [List-of [List-of Integer ]]] -> Progress-Data
;
; Create a new `Progress-Data` struct. Optionally, you can pass in a list of
; lists where each sublist contains values for recursion level, board index and
; guess. The latter value may be `BEFORE-GUESSES`.
(define (make-progress-data [records-data '()])
  (define progress-data (Progress-Data (gvector)))
  (for ([record-data records-data])
    (apply add-progress-record! progress-data record-data))
  progress-data)

; Progress-Data -> string
;
; Convert progress data to a string. This is mostly used for debugging.
(define (progress-data->string progress-data)
  (string-join
    (for/list ([record (in-gvector (progress-data-records progress-data))])
      (format "~a,~a,~a" (record-recursion-level record)
                         (record-board-index record)
                         (record-guess record)))
    "\n"))

; --------------------------------------------------------------------------------
; Generate recursion statistics

; We could import this from a common module, but let's keep things simple for
; now.
(define BOARD-SIZE 81)

; Vector Integer -> Void
;
; In `vector`, increment integer at position `pos` by 1.
(define (vector-inc! vec pos)
  (vector-set! vec pos (add1 (vector-ref vec pos))))

; Progress-Data -> [Vector-of Integer]
;
; Return histogram data mapping recursion level to the number of times that
; recursion level was entered.
(define (recursion-stats progress-data)
  (define records (progress-data-records progress-data))
  ; We can have at most as many recursion levels as there are cells on the
  ; board. However, add 1 to account for recursion levels starting at 1, not 0.
  (define stats (make-vector (add1 BOARD-SIZE) 0))
  (for ([record (in-gvector records)]
        #:when (eq? (record-guess record) BEFORE-GUESSES))
    (define recursion-level (record-recursion-level record))
    (vector-inc! stats recursion-level))
  stats)

; --------------------------------------------------------------------------------
; Helper function for text and graph output

; Progress-Data -> [Hash-of Integer [Integer | #f]
;
; Return a hash mapping a cell index to the cell index of the successor. For
; the last index, use `#f` as successor.
(define (cell-successors-hash progress-data)
  (define guessable-board-indices
    (sort
      (remove-duplicates
        (for/list ([record (progress-data-records progress-data)])
          (record-board-index record)))
      <))
  (define last-list-index (sub1 (length guessable-board-indices)))
  (define successors-hash (make-hash))
  (for ([list-index (in-naturals)]
        [board-index guessable-board-indices])
    (hash-set!
      successors-hash
      board-index
      ; Use `<` because we need special handling for the last list index.
      (if (< list-index last-list-index)
        (list-ref guessable-board-indices (add1 list-index))
        #f)))
  successors-hash)

; --------------------------------------------------------------------------------
; Nested text output from progress data

; Progress-Data Output-Port -> String
;
; Return a multi-line string representing the solver progress. Each line is formatted as
;
;   board-index --guess--> next-board-index
;
; with lines indented by recursion level. The line means that the solver could make a
; valid guess `guess` at `board-index` that made it feasible to continue with guesses at
; index `next-board-index`.
(define (progress-data->text-file progress-data output-port)
  (define successors-hash (cell-successors-hash progress-data))
  (parameterize ([current-output-port output-port])
    (for ([record (progress-data-records progress-data)]
          #:unless (eq? (record-guess record) BEFORE-GUESSES))
      (define indent (make-string (sub1 (record-recursion-level record))
                                  #\space))
      (define from-board-index (record-board-index record))
      (define label (record-guess record))
      (define to-board-index (hash-ref successors-hash from-board-index))
      (displayln
        (format "~a~a --~a--> ~a" indent
                                  from-board-index
                                  label
                                  (or to-board-index "done"))))))

; --------------------------------------------------------------------------------
; Solution tree generation from the progress data

; [List-of Record] -> [List-of [List-of Record]]
;
; Partition the records so that each partition starts with the same recursion
; level as the first record in `records`. In other words, each partition
; describes a subset of `records` where the top-level iteration tried a
; different guess.
;
; "Dummy records", that is with the guess set to `BEFORE-GUESSES`, aren't
; included in the result.
;
; Example: If the `records` are (indented by recursion level for clarity)
;
; 8,11,2
;  9,15,3
;  9,15,7
;   10,16,3
; 8,11,7
;  9,15,3
;   10,16,2
;
; the partitions are the two record lists
;
; 8,11,2
;  9,15,3
;  9,15,7
;   10,16,3
;
; and
;
; 8,11,7
;  9,15,3
;   10,16,2
(define (partition-records records)
  (cond
    [(null? records)
     '()]
    [else
     (define root-recursion-level (record-recursion-level (car records)))
     (for/fold ([partitions '()]
                #:result
                ; Reverse inner and outer list(s) because we built the
                ; lists "from back to front."
                (reverse (map reverse partitions)))
               ([record records])
       (cond
         [(eq? (record-guess record) BEFORE-GUESSES)
          ; Skip records without a "real" guess.
          partitions]
         [else
          (define recursion-level (record-recursion-level record))
          (cond
            [(= recursion-level root-recursion-level)
             ; Start new partition.
             (cons (list record)
                   partitions)]
            [else
             ; Add record to first list in `partitions`.
             (cons (cons record (car partitions))
                   (cdr partitions))])]))]))

; [List-of Record] [Hash-of Integer [Integer | #f]] -> [List-of Node]
;
; Return a list of - possibly recursive - child nodes. The child node labels
; and the edge labels leading to them are derived from `records-list`.
(define (node-child-nodes records-list successors-hash)
  (define partitions (partition-records records-list))
  (for/list ([partition partitions])
    (define first-record (car partition))
    (Node
      (~a (or (hash-ref successors-hash (record-board-index first-record))
              "done"))
      (~a (record-guess first-record))
      (node-child-nodes (cdr partition) successors-hash))))

; Progress-Data -> Node
;
; Return a node including its child nodes, representing the graph over all of
; the solution process.
(define (progress-data->node progress-data)
  (define records-list (gvector->list (progress-data-records progress-data)))
  (define successors-hash (cell-successors-hash progress-data))
  (Node
    (~a (record-board-index (car records-list)))
    ""
    (node-child-nodes records-list successors-hash)))

; --------------------------------------------------------------------------------

(module+ test
  (require
    rackunit
    rackunit/text-ui)

  (run-tests
    (test-suite "Progress data"
      (test-case
        "Empty progress data"
        (define empty-progress-data (make-progress-data))
        (define records (progress-data-records empty-progress-data))
        (check-equal? (gvector-count records) 0))
      (test-case
        "Add a few records"
        (define progress-data (make-progress-data))
        (add-progress-record! progress-data 1 11 3)
        (add-progress-record! progress-data 2 15 BEFORE-GUESSES)
        (add-progress-record! progress-data 2 15 3)
        (define records (progress-data-records progress-data))
        (check-equal? (gvector-count records) 3)
        (for ([record records]
              [expected-record (list (Record 1 11 3)
                                     (Record 2 15 BEFORE-GUESSES)
                                     (Record 2 15 3))])
          (check-equal? record expected-record)))
      (test-case
        "Add a few records with `make-progress-data`"
        (define progress-data
          (make-progress-data
            `((1 11 3)
              (2 15 ,BEFORE-GUESSES)
              (2 15 3))))
        (define records (progress-data-records progress-data))
        (check-equal? (gvector-count records) 3)
        (for ([record records]
              [expected-record (list (Record 1 11 3)
                                     (Record 2 15 BEFORE-GUESSES)
                                     (Record 2 15 3))])
          (check-equal? record expected-record)))
      (test-case
        "Recursion stats"
        ; Create progress data.
        (define progress-data
          (make-progress-data
            `((1 11 3)
              (2 15 ,BEFORE-GUESSES)
              (2 15 3)
              (3 18 ,BEFORE-GUESSES)
              (3 18 5)
              (2 15 ,BEFORE-GUESSES)
              (2 15 8))))
        (define stats (recursion-stats progress-data))
        ; Create stats for comparison.
        (define expected-stats (make-vector (add1 BOARD-SIZE) 0))
        (vector-set! expected-stats 2 2)
        (vector-set! expected-stats 3 1)
        (check-equal? stats expected-stats))
      (test-case
        "Text output"
        ; Create progress data.
        (define progress-data
          (make-progress-data
            `((1 11 3)
              (2 15 ,BEFORE-GUESSES)
              (2 15 3)
              (2 15 ,BEFORE-GUESSES)
              (2 15 8)
              (3 18 ,BEFORE-GUESSES)
              (3 18 5))))
        (define output-port (open-output-string))
        (progress-data->text-file progress-data output-port)
        (check-equal?
          (get-output-string output-port)
          "11 --3--> 15\n 15 --3--> 18\n 15 --8--> 18\n  18 --5--> done\n"))
  ))

  (run-tests
    (test-suite "Record partitioning"
      (test-equal?
        "Empty partitions"
        (partition-records '())
        '())
      (test-equal?
        "Partition example"
        (partition-records
          (list (make-record 8 11 2)
                (make-record 9 15 3)
                (make-record 9 15 7)
                (make-record 10 16 3)
                (make-record 8 11 7)
                (make-record 9 15 3)
                (make-record 10 16 2)))
        (list
          (list (make-record 8 11 2)
                (make-record 9 15 3)
                (make-record 9 15 7)
                (make-record 10 16 3))
          (list (make-record 8 11 7)
                (make-record 9 15 3)
                (make-record 10 16 2))))
      (test-equal?
        "Skip dummy records"
        (partition-records
          (list (make-record 8 11 BEFORE-GUESSES)
                (make-record 8 11 2)
                (make-record 9 15 BEFORE-GUESSES)
                (make-record 9 15 3)
                (make-record 9 15 7)
                (make-record 10 16 BEFORE-GUESSES)
                (make-record 10 16 3)
                (make-record 8 11 7)
                (make-record 9 15 BEFORE-GUESSES)
                (make-record 9 15 3)
                (make-record 10 16 BEFORE-GUESSES)
                (make-record 10 16 2)))
        (list
          (list (make-record 8 11 2)
                (make-record 9 15 3)
                (make-record 9 15 7)
                (make-record 10 16 3))
          (list (make-record 8 11 7)
                (make-record 9 15 3)
                (make-record 10 16 2))))
  ))
)
